import json
from pathlib import Path

from resolwe import process
from resolwe.process import (
    Cmd,
    DataField,
    DirField,
    FileField,
    FloatField,
    GroupField,
    IntegerField,
    JsonField,
    ListField,
    Persistence,
    Process,
    SchedulingClass,
    StringField,
    UrlField,
)
from resolwe.process.models import Collection, Data, Entity
from resolwe.process.models import Process as ProcessM


class EntityProcess(Process):
    slug = "entity-process"
    name = "Entity process"
    data_name = "Data with entity"
    version = "1.0.0"
    process_type = "data:entity"
    entity = {
        "type": "sample",
    }

    class Output:
        list_string = ListField(StringField(), label="My list")
        optional = StringField("Optional output", required=False)

    def run(self, inputs, outputs):
        outputs.list_string = ["foo", "bar"]


class PythonProcess(Process):
    """This is a process description."""

    slug = "test-python-process"
    name = "Test Python Process"
    version = "0.1.2"
    process_type = "data:python"
    category = "analyses"
    scheduling_class = SchedulingClass.BATCH
    persistence = Persistence.CACHED
    data_name = "Foo: {{input_data | name}}"
    entity = {
        "type": "sample",
        "descriptor_schema": "sample",
        "input": "input_data",
    }
    requirements = {
        "expression-engine": "jinja",
        "executor": {
            "docker": {
                "image": "public.ecr.aws/s4q6j6e8/resolwe/base:ubuntu-20.04",
            }
        },
    }

    class Input:
        """Input fields."""

        my_field = StringField(label="My field")
        my_list = ListField(StringField(), label="My list")
        input_data = DataField("test:save", label="My input data")
        input_entity_data = DataField("entity", label="My entity data")
        bar = DataField(data_type="test:save", label="My bar", relation_type="group")
        baz = DataField(
            data_type="test:save",
            required=False,
            label="My bar",
            relation_type="group",
            relation_npartitions=1,
        )
        baz_list = ListField(
            DataField(data_type="test:save"),
            required=False,
            label="My baz list",
            relation_type="group",
            relation_npartitions=1,
        )
        url = UrlField(UrlField.DOWNLOAD, label="My URL")
        integer = IntegerField(label="My integer")
        my_float = FloatField(label="My float")
        my_json = JsonField(label="Blah blah")
        my_optional = StringField(
            label="Optional", required=False, default="default value"
        )
        my_optional_no_default = StringField(
            label="Optional no default", required=False
        )

        class MyGroup:
            foo = IntegerField(label="Foo")
            bar = StringField(label="Bar")
            group_optional_no_default = StringField(
                label="Group optional no default", required=False
            )

        my_group = GroupField(MyGroup, label="My group")

    class Output:
        string_output = StringField(label="My string output")
        list_string_output = ListField(StringField(), label="My list string output")
        file_output = FileField(label="My output")
        list_file_output = ListField(FileField(), label="My list output")
        dir_output = DirField(label="My output")
        input_data_name = StringField(label="Input data name")
        input_entity_name = StringField(label="Input entity name")
        docker_image = StringField(label="Docker image")

    def run(self, inputs, outputs):
        print("All inputs are:", inputs)
        print("Input data:", inputs.input_data)
        print("Input data ID:", inputs.input_data.id)
        print("Input data file output:", inputs.input_data.output.saved_file.path)
        print("Input data type:", inputs.input_data.type)
        print("Input data descriptor:", inputs.input_data.descriptor)
        print("Group bar:", inputs.my_group.bar)
        print("Group foo:", inputs.my_group.foo)
        print("Entity name of the input:", inputs.input_entity_data.entity.name)
        print("Docker image:", self.requirements.executor.docker.image)

        if inputs.my_optional:
            print("My optional:", inputs.my_optional)

        if inputs.my_optional_no_default:
            raise AttributeError("inputs.my_optional_no_default should not exist.")

        if inputs.my_group.group_optional_no_default:
            raise AttributeError(
                "inputs.my_group.group_optional_no_default should not exist."
            )

        if inputs.input_entity_data.output.optional:
            raise AttributeError("inputs.list_string_output.optional should not exist.")

        try:
            inputs.invalid_input
        except AttributeError as err:
            if "Inputs have no field invalid_input" in str(err):
                pass

        try:
            inputs.input_entity_data.output.invalid_field
        except AttributeError as err:
            if "DataField has no member invalid_field" in str(err):
                pass

        bar = Cmd["ls"]["-l", "-a", "/"] | Cmd["grep"]["python"]
        print("hello world:\n", bar())

        Cmd["mkdir"]["test"]()
        (Cmd["echo"]['"Some content"'] > "test/testfile.txt")()
        (Cmd["echo"]['"Some more content"'] > "testfile2.txt")()

        outputs.file_output = "test/testfile.txt"
        outputs.list_file_output = ["test/testfile.txt", "testfile2.txt"]
        outputs.dir_output = "test/"
        outputs.input_data_name = inputs.input_data.name
        outputs.input_entity_name = inputs.input_entity_data.entity.name
        outputs.docker_image = self.requirements.executor.docker.image
        outputs.string_output = "OK"
        outputs.list_string_output = ["foo", "bar"]


class PythonProcessGroup(Process):
    """Process with Group fields."""

    slug = "test-python-process-group-field"
    name = "Test Python Process for GroupField"
    version = "0.1.2"
    process_type = "data:python:group"

    class Input:
        """Input fields."""

        class MyGroup:
            foo = IntegerField(label="Foo", required=False, default=42)
            bar = StringField(label="Bar", required=False)

        class MyGroup2:
            foo = IntegerField(label="Foo", required=False)

        my_group = GroupField(MyGroup, label="My group")
        my_group2 = GroupField(
            MyGroup2, label="My group2 that has all elements without defaults."
        )

    class Output:
        """Output fields."""

        out_foo = IntegerField(label="Foo.", required=False)
        out_bar = StringField(label="Bar.", required=False)
        out_foo2 = IntegerField(label="Foo2.", required=False)

    def run(self, inputs, outputs):
        if inputs.my_group.foo:
            outputs.out_foo = inputs.my_group.foo
        if inputs.my_group.bar:
            outputs.out_bar = inputs.my_group.bar
        if inputs.my_group2.foo:
            outputs.out_foo2 = inputs.my_group2.foo


class PythonProcess2(process.Process):
    """Inherit from 'module.Class'."""

    slug = "test-python-process-2"
    name = "Test Python Process 2"
    version = "0.0.1"
    process_type = "data:python"

    def run(self, inputs, outputs):
        pass


class ErrorProcess(Process):
    slug = "test-python-process-error"
    name = "Test Python Process Error"
    version = "0.0.1"
    process_type = "data:python:error"

    def run(self, inputs, outputs):
        raise ValueError("Value error in ErrorProcess")


class AnnotateProcess(Process):
    slug = "test-python-process-annotate-entity"
    name = "Test Python Process Annotate Entity"
    version = "0.0.1"
    process_type = "data:python:annotate"
    entity = {
        "type": "sample",
    }

    def run(self, inputs, outputs):
        annotations = {"general.description": "desc", "general.species": "Valid"}
        self.update_entity_descriptor(annotations)


class FileProcess(Process):
    slug = "test-python-process-file"
    name = "Test Python Process File"
    version = "0.0.1"
    process_type = "data:python:file"

    class Input:
        """Input fields."""

        src = FileField(label="Input file")

    class Output:
        """Input fields."""

        dst = FileField(label="Output file")

    def run(self, inputs, outputs):
        file_name = inputs.src.import_file()
        outputs.dst = file_name


class RequirementsProcess(Process):
    slug = "test-python-process-requirements"
    name = "Test Python Process Requirements"
    version = "0.0.1"
    process_type = "data:python:requirements"
    requirements = {
        "resources": {
            "cores": 2,
            "memory": 4096,
        },
    }

    class Output:
        """Input fields."""

        cores = IntegerField(label="Cores")
        memory = IntegerField(label="Memory")

    def run(self, inputs, outputs):
        outputs.cores = self.requirements["resources"]["cores"]
        outputs.memory = self.requirements["resources"]["memory"]
        outputs.cores = self.requirements["resources"]["cores"]

        print("Cores:", outputs.cores)
        print("Memory:", outputs.memory)


class ProcessWithWorkflowInput(Process):
    slug = "process-with-workflow-input"
    name = "Process with workflow input"
    version = "1.0.0"
    process_type = "data:test"

    class Input:
        """Input fields."""

        data = DataField(data_type="", label="Data.")

    def run(self, inputs, outputs):
        pass


class ProcessWithChoicesInput(Process):
    slug = "process-with-choices-input"
    name = "Process with choices input"
    version = "1.0.0"
    process_type = "data:test"

    class Input:
        """Input fields."""

        string_input = StringField(
            label="Input field",
            allow_custom_choice=True,
            choices=[
                ("foo", "foo"),
                ("bar", "bar"),
            ],
        )

    class Output:
        """Output fields."""

        string_output = StringField(label="Output field")

    def run(self, inputs, outputs):
        outputs.string_output = inputs.string_input


class RelationsProcess(Process):
    slug = "test-process-relations"
    name = "Test Python Process relations"
    version = "0.0.1"
    process_type = "data:python:relations"
    requirements = {
        "relations": [{"type": "series"}],
    }

    class Input:
        """Input fields."""

        data = ListField(DataField(data_type=""), label="Data.")

    class Output:
        """Output fields."""

        relation_id = IntegerField(label="Relation id")
        relation_type = StringField(label="Relation type")
        relation_ordered = StringField(label="Relation ordering")
        relation_category = StringField(label="Relation category")
        relation_unit = StringField(label="Relation unit")
        relation_partition_label = StringField(label="Relation partition label")
        relation_partition_position = IntegerField(label="Relation partition label")

    def run(self, inputs, outputs):
        # Access relation attributes
        outputs.relation_id = inputs.data[0].relations[0].id
        outputs.relation_type = inputs.data[0].relations[0].type
        outputs.relation_ordered = str(inputs.data[0].relations[0].ordered)
        outputs.relation_category = inputs.data[0].relations[0].category
        outputs.relation_unit = inputs.data[0].relations[0].unit
        # Access relation partition attributes
        outputs.relation_partition_label = (
            inputs.data[0].relations[0].partitions[0].label
        )
        outputs.relation_partition_position = (
            inputs.data[0].relations[0].partitions[0].position
        )


class DataNameProcess(Process):
    slug = "data-name-process"
    name = "Data name process"
    data_name = "{{ data_input | name | default('?') }}"
    version = "1.0.0"
    process_type = "data:name"
    requirements = {"expression-engine": "jinja"}

    class Input:
        data_input = DataField(data_type="", label="Data input")

    class Output:
        name = StringField(label="Name")

    def run(self, inputs, outputs):
        outputs.name = self.name


class CreateCollection(Process):
    slug = "create-collection"
    name = "Create collection"
    data_name = "{{ data_input | name | default('?') }}"
    version = "1.0.0"
    process_type = "data:name"
    requirements = {"expression-engine": "jinja"}

    class Input:
        collection_name = StringField(label="Collection name")

    def run(self, inputs, outputs):
        Collection.create(name=inputs.collection_name)


class GetCollection(Process):
    slug = "get-collection"
    name = "Get collection"
    data_name = "{{ data_input | name | default('?') }}"
    version = "1.0.0"
    process_type = "data:name"
    requirements = {"expression-engine": "jinja"}

    class Input:
        collection_name = StringField(label="Collection name")

    class Output:
        collection_slug = StringField(label="Collection slug")

    def run(self, inputs, outputs):
        collection = Collection.get(name=inputs.collection_name)
        outputs.collection_slug = collection.slug


class FilterCollection(Process):
    slug = "filter-collection"
    name = "Filter collection"
    data_name = "{{ data_input | name | default('?') }}"
    version = "1.0.0"
    process_type = "data:name"
    requirements = {"expression-engine": "jinja"}

    class Input:
        collection_name = StringField(label="Collection name")

    class Output:
        number_of_collections = IntegerField(label="Number of collections")

    def run(self, inputs, outputs):
        collections = Collection.filter(name=inputs.collection_name)
        outputs.number_of_collections = len(collections)


class CreateData(Process):
    slug = "create-data"
    name = "Create data object"
    data_name = "{{ data_input | name | default('?') }}"
    version = "1.0.0"
    process_type = "data:name"
    requirements = {"expression-engine": "jinja"}

    class Input:
        data_name = StringField(label="Data name")
        collection_name = StringField(label="Collection name")

    def run(self, inputs, outputs):
        process = ProcessM.get(slug="create-collection")
        Data.create(
            process=process,
            name=inputs.data_name,
            input={"collection_name": inputs.collection_name},
        )


class CreateEntityTags(Process):
    slug = "assign-entity-tags"
    name = "Create entity tags"
    data_name = "{{ data_input | name | default('?') }}"
    version = "1.0.0"
    process_type = "data:name"
    requirements = {"expression-engine": "jinja"}

    class Input:
        data_name = StringField(label="Data name")
        sample_name = StringField(label="Collection name")
        tags = ListField(StringField(), label="My tags")

    def run(self, inputs, outputs):
        sample = Entity.create(name=inputs.sample_name)
        sample.tags = inputs.tags
        self.data.name = inputs.data_name
        self.data.entity = sample


class ChangeEntityName(Process):
    slug = "change-entity-name"
    name = "Rename entity"
    data_name = "{{ data_input | name | default('?') }}"
    version = "1.0.0"
    process_type = "data:name"
    requirements = {"expression-engine": "jinja"}

    class Input:
        entity_id = IntegerField(label="Entity id")
        entity_name = StringField(label="New entity name")

    def run(self, inputs, outputs):
        entity = Entity.get(pk=inputs.entity_id)
        entity.name = inputs.entity_name


class TestStorage(Process):
    slug = "storage-objects-test"
    name = "Test working with storage objects"
    data_name = "{{ data_input | name | default('?') }}"
    version = "1.0.0"
    process_type = "data:name"
    requirements = {"expression-engine": "jinja"}

    class Output:
        output_string = JsonField(label="Output string")
        output_file = JsonField(label="Output file")

    def run(self, inputs, outputs):
        # Test update of storage objects.
        outputs.output_string = ["first", "valid", "json"]
        outputs.output_string = ["valid", "json"]
        # Test storage objects are read from file (it it exists).
        json_filename = "json_file.out"
        Path(json_filename).write_text(json.dumps(["valid", "json", "file"]))
        outputs.output_file = json_filename



"""Reslowe process model."""
from django.conf import settings
from django.contrib.postgres.indexes import GinIndex
from django.core.validators import RegexValidator
from django.db import models

from .base import BaseModel


class Process(BaseModel):
    """Postgres model for storing processes."""

    class Meta(BaseModel.Meta):
        """Process Meta options."""

        permissions = (
            ("view_process", "Can view process"),
            ("share_process", "Can share process"),
            ("owner_process", "Is owner of the process"),
        )

        indexes = [
            models.Index(name="idx_process_name", fields=["name"]),
            models.Index(name="idx_process_slug", fields=["slug"]),
            models.Index(name="idx_process_type", fields=["type"]),
            GinIndex(
                name="idx_process_type_trgm",
                fields=["type"],
                opclasses=["gin_trgm_ops"],
            ),
        ]

    #: raw persistence
    PERSISTENCE_RAW = "RAW"
    #: cached persistence
    PERSISTENCE_CACHED = "CAC"
    #: temp persistence
    PERSISTENCE_TEMP = "TMP"
    PERSISTENCE_CHOICES = (
        (PERSISTENCE_RAW, "Raw"),
        (PERSISTENCE_CACHED, "Cached"),
        (PERSISTENCE_TEMP, "Temp"),
    )

    SCHEDULING_CLASS_INTERACTIVE = "IN"
    SCHEDULING_CLASS_BATCH = "BA"
    SCHEDULING_CLASS_CHOICES = (
        (SCHEDULING_CLASS_INTERACTIVE, "Interactive"),
        (SCHEDULING_CLASS_BATCH, "Batch"),
    )

    #: data type
    type = models.CharField(
        max_length=100,
        validators=[
            RegexValidator(
                regex=r"^data:[a-z0-9:]+:$",
                message="Type may be alphanumerics separated by colon",
                code="invalid_type",
            )
        ],
    )

    #: category
    category = models.CharField(
        max_length=200,
        default="Other:",
        validators=[
            RegexValidator(
                regex=r"^([a-zA-Z0-9]+[:\-])*[a-zA-Z0-9]+:$",
                message="Category may be alphanumerics separated by colon",
                code="invalid_category",
            )
        ],
    )

    persistence = models.CharField(
        max_length=3, choices=PERSISTENCE_CHOICES, default=PERSISTENCE_RAW
    )
    """
    Persistence of :class:`~resolwe.flow.models.Data` objects created
    with this process. It can be one of the following:

    - :attr:`PERSISTENCE_RAW`
    - :attr:`PERSISTENCE_CACHED`
    - :attr:`PERSISTENCE_TEMP`

    .. note::

        If persistence is set to ``PERSISTENCE_CACHED`` or
        ``PERSISTENCE_TEMP``, the process must be idempotent.

    """

    #: designates whether this process should be treated as active
    is_active = models.BooleanField("active", default=True)

    #: detailed description
    description = models.TextField(default="")

    #: template for name of Data object created with Process
    data_name = models.CharField(max_length=200, null=True, blank=True)

    input_schema = models.JSONField(blank=True, default=list)
    """
    process input schema (describes input parameters, form layout **"Inputs"** for :attr:`Data.input`)

    Handling:

    - schema defined by: *dev*
    - default by: *user*
    - changable by: *none*

    """

    output_schema = models.JSONField(blank=True, default=list)
    """
    process output schema (describes output JSON, form layout **"Results"** for :attr:`Data.output`)

    Handling:

    - schema defined by: *dev*
    - default by: *dev*
    - changable by: *dev*

    Implicitly defined fields (by
    :func:`resolwe.flow.management.commands.register` or
    ``resolwe.flow.executors.run.BaseFlowExecutor.run`` or its
    derivatives):

    - ``progress`` of type ``basic:float`` (from 0.0 to 1.0)
    - ``proc`` of type ``basic:group`` containing:

      - ``stdout`` of type ``basic:text``
      - ``rc`` of type ``basic:integer``
      - ``task`` of type ``basic:string`` (celery task id)
      - ``worker`` of type ``basic:string`` (celery worker hostname)
      - ``runtime`` of type ``basic:string`` (runtime instance hostname)
      - ``pid`` of type ``basic:integer`` (process ID)

    """

    entity_type = models.CharField(max_length=100, null=True, blank=True)
    """
    Automatically add :class:`~resolwe.flow.models.Data` object created
    with this process to an :class:`~resolwe.flow.models.Entity` object
    representing a data-flow. If all input ``Data`` objects belong to
    the same entity, add newly created ``Data`` object to it, otherwise
    create a new one.
    """

    entity_descriptor_schema = models.CharField(max_length=100, null=True, blank=True)
    """
    Slug of the descriptor schema assigned to the Entity created with
    :attr:`~resolwe.flow.models.Process.entity_type`.
    """

    entity_input = models.CharField(max_length=100, null=True, blank=True)
    """
    Limit the entity selection in
    :attr:`~resolwe.flow.models.Process.entity_type` to a single input.
    """

    entity_always_create = models.BooleanField(default=False)
    """
    Create new entity, regardless of ``entity_input`` or
    ``entity_descriptor_schema`` fields.
    """

    run = models.JSONField(default=dict)
    """
    process command and environment description for internal use

    Handling:

    - schema defined by: *dev*
    - default by: *dev*
    - changable by: *dev*

    """

    requirements = models.JSONField(default=dict)
    """
    process requirements
    """

    scheduling_class = models.CharField(
        max_length=2, choices=SCHEDULING_CLASS_CHOICES, default=SCHEDULING_CLASS_BATCH
    )
    """
    process scheduling class
    """

    def get_resource_limits(self):
        """Get the core count and memory usage limits for this process.

        :return: A dictionary with the resource limits, containing the
            following keys:

            - ``memory``: Memory usage limit, in MB. Defaults to 4096 if
              not otherwise specified in the resource requirements.
            - ``cores``: Core count limit. Defaults to 1.
            - ``storage``: Size (in gibibytes) of temporary volume used for
              processing in kubernetes. Defaults to 200.

        :rtype: dict
        """
        # Get limit defaults and overrides.
        limit_defaults = getattr(settings, "FLOW_PROCESS_RESOURCE_DEFAULTS", {})
        limit_overrides = getattr(settings, "FLOW_PROCESS_RESOURCE_OVERRIDES", {})

        limits = {}

        resources = self.requirements.get("resources", {})

        limits["cores"] = int(resources.get("cores", 1))

        max_cores = getattr(settings, "FLOW_PROCESS_MAX_CORES", None)
        if max_cores:
            limits["cores"] = min(limits["cores"], max_cores)

        memory = limit_overrides.get("memory", {}).get(self.slug, None)
        if memory is None:
            memory = int(
                resources.get(
                    "memory",
                    # If no memory resource is configured, check settings.
                    limit_defaults.get("memory", 4096),
                )
            )
        limits["memory"] = memory

        limits["storage"] = limit_overrides.get("storage", {}).get(
            self.slug, int(resources.get("storage", limit_defaults.get("storage", 200)))
        )

        return limits



"""Python process communicator."""
import os
import socket
from pathlib import Path
from typing import Any, Optional, Type

from .socket_utils import Message, Response, receive_data, send_data


class Singleton:
    """Decorator class for singleton."""

    def __init__(self, klass: Type):
        """Initialize."""
        self.klass = klass
        self.instance = None

    def __call__(self, *args, **kwargs):
        """Override standard __call__ method."""
        if self.instance is None:
            self.instance = self.klass(*args, **kwargs)
        return self.instance


@Singleton
class PythonProcessCommunicator:
    """Base communicator for communicating with communication container."""

    __instance: Optional[
        "PythonProcessCommunicator"
    ] = None  #  A single instance of this class

    def __init__(self, _socket: socket.SocketType):
        """Initialize."""
        self._socket = _socket
        self.encoder = None

    def send_command(
        self,
        command_name: str,
        data: Any,
    ) -> Response:
        """Send data and return the response.

        :raises AssertionError: on error.
        """
        command = Message.command(command_name, data)
        send_data(self._socket, command.to_dict(), encoder=self.encoder)
        received = receive_data(self._socket)
        assert received is not None
        return Response.from_dict(received)

    def __getattr__(self, name: str):
        """Call arbitrary command with 'com.command(args)' syntax.

        When attribute is requested that is not known the method is returned
        that will call the ``send_command`` method with the given arguments
        and returt the ``message_data`` of the received answer.
        """

        def call_command(*args):
            if len(args) == 1:
                args = args[0]
            return self.send_command(name, args).message_data

        if name.startswith("_"):
            return None
        else:
            return call_command


def get_communicator():
    """Create and return a communicator instance."""
    SOCKET_TIMEOUT: Optional[int] = None
    if "SOCKET_TIMEOUT" == os.environ:
        SOCKET_TIMEOUT = int(os.environ["SOCKET_TIMEOUT"])
    SOCKETS_PATH = Path(os.environ.get("SOCKETS_VOLUME", "/sockets"))
    PROCESSING_CONTAINER_SOCKET = SOCKETS_PATH / os.environ.get(
        "SCRIPT_SOCKET", "_socket2.s"
    )

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.connect(os.fspath(PROCESSING_CONTAINER_SOCKET))
    sock.settimeout(SOCKET_TIMEOUT)
    return PythonProcessCommunicator(sock)


communicator = None
if "RUNNING_IN_CONTAINER" in os.environ:
    communicator = get_communicator()



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-09-25 08:51
from __future__ import unicode_literals

from django.db import migrations

from resolwe.flow.utils import iterate_fields


def update_dependency_kinds(apps, schema_editor):
    """Update historical dependency kinds as they may be wrong."""
    DataDependency = apps.get_model("flow", "DataDependency")
    for dependency in DataDependency.objects.all():
        # Assume dependency is of subprocess kind.
        dependency.kind = "subprocess"

        # Check child inputs to determine if this is an IO dependency.
        child = dependency.child
        parent = dependency.parent

        for field_schema, fields in iterate_fields(
            child.input, child.process.input_schema
        ):
            name = field_schema["name"]
            value = fields[name]

            if field_schema.get("type", "").startswith("data:"):
                if value == parent.pk:
                    dependency.kind = "io"
                    break
            elif field_schema.get("type", "").startswith("list:data:"):
                for data in value:
                    if value == parent.pk:
                        dependency.kind = "io"
                        break

        dependency.save()


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0004_data_dependency_2"),
    ]

    operations = [migrations.RunPython(update_dependency_kinds)]



"""Descriptor schema viewset."""
from rest_framework import mixins, viewsets

from resolwe.flow.filters import DescriptorSchemaFilter
from resolwe.flow.models import DescriptorSchema
from resolwe.flow.serializers import DescriptorSchemaSerializer
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin


class DescriptorSchemaViewSet(
    mixins.RetrieveModelMixin,
    mixins.ListModelMixin,
    ResolwePermissionsMixin,
    viewsets.GenericViewSet,
):
    """API view for :class:`DescriptorSchema` objects."""

    queryset = DescriptorSchema.objects.all().select_related("contributor")
    serializer_class = DescriptorSchemaSerializer
    permission_classes = (get_permissions_class(),)
    filterset_class = DescriptorSchemaFilter
    ordering_fields = ("id", "created", "modified", "name", "version")
    ordering = ("id",)



"""Resolwe Python process support."""
import logging
import sys

from .communicator import communicator
from .descriptor import Persistence, ProcessDescriptor, SchedulingClass, ValidationError
from .fields import (
    BooleanField,
    DataField,
    DateField,
    DateTimeField,
    DirField,
    FileField,
    FileHtmlField,
    FloatField,
    GroupField,
    IntegerField,
    JsonField,
    ListField,
    SecretField,
    StringField,
    TextField,
    UrlField,
)
from .models import Collection, Data, Entity
from .runtime import Process

try:
    from plumbum import local as Cmd

    # Log plumbum commands to standard output.

    formatter = logging.Formatter(
        "%(asctime)s - %(levelname)s - %(name)s[%(process)s]: %(message)s"
    )

    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(formatter)

    plumbum_logger = logging.getLogger("plumbum.local")
    plumbum_logger.setLevel(logging.DEBUG)
    plumbum_logger.addHandler(handler)


except ImportError:
    # Generate a dummy class that defers the ImportError to a point where
    # the package actually tries to be used.
    class MissingPlumbumCmd:
        """Missing plumbum command dummy class."""

        def __getitem__(self, cmd):
            """Raise ImportError exception."""
            raise ImportError("Missing 'plumbum' Python package.")

    Cmd = MissingPlumbumCmd()


__all__ = (
    "BooleanField",
    "Cmd",
    "DataField",
    "DateField",
    "DateTimeField",
    "DirField",
    "FileField",
    "FileHtmlField",
    "FloatField",
    "GroupField",
    "IntegerField",
    "JsonField",
    "ListField",
    "Persistence",
    "Process",
    "ProcessDescriptor",
    "SchedulingClass",
    "SecretField",
    "StringField",
    "ValidationError",
    "TextField",
    "UrlField",
    "Data",
    "Entity",
    "Collection",
    "communicator",
)


# XXX: ImportedFormat is available from resolwe-runtime-utils v2.0.0. Docker
# images with older version of resolwe-runtime-utils should skip the import.
try:
    from resolwe_runtime_utils import ImportedFormat

    __all__ += (ImportedFormat,)
except ImportError:
    pass



"""Python process runner utility."""
import inspect
import logging
import os
import sys
import types
from typing import Dict, Type

from .communicator import communicator
from .descriptor import ValidationError
from .models import Data
from .runtime import Process

# Id of the Data object we are processing.
DATA_ID = int(os.getenv("DATA_ID", "-1"))
logger = logging.getLogger(__name__)


if __name__ == "__main__":
    assert communicator is not None

    # Get the data object.
    data = Data(DATA_ID)
    module = types.ModuleType("python_program")
    python_program = communicator.get_python_program()
    exec(python_program, module.__dict__)

    # Mapping between the process slug and the class containing the process
    # definition.
    processes: Dict[str, Type[Process]] = {}
    for variable in dir(module):
        value = getattr(module, variable)
        if (
            value == Process
            or not inspect.isclass(value)
            or not issubclass(value, Process)
            or "_abstract" in value.__dict__
        ):
            continue
        processes[value._meta.metadata.slug] = value

    try:
        process = processes[data.process.slug]
    except KeyError:
        print("Found the following processes in module '{}':".format(processes))
        print("")
        for slug, process in processes.items():
            print("  {} ({})".format(slug, getattr(process, "name")))
        print("")

        print("ERROR: Unable to find process '{}'.".format(data.process.slug))
        sys.exit(1)

    # Start the process.
    try:
        process(data).start()
    except ValidationError as error:
        print("ERROR: Output field validation failed: {}".format(error.args[0]))
        sys.exit(1)



import os

from django.db import migrations, connection
from django.conf import settings


def collect(base_dir):
    """Collect all files in base_dir."""
    collected_dirs = []
    collected_files = []
    for root, _, files in os.walk(os.path.join(base_dir)):
        # Make sure directory path ends with '/', do not include './'
        if root != base_dir:
            collected_dirs.append(os.path.join(os.path.relpath(root, base_dir), ""))
        for file_ in files:
            real_path = os.path.join(root, file_)
            file_size = os.path.getsize(real_path)
            collected_files.append((os.path.relpath(real_path, base_dir), file_size))
    return collected_files, collected_dirs


def create_new_storage_locations(apps, schema_editor):
    """Create FileStorage objects and StorageLocation from each
    DataLocation object. Make sure that ids of DataLocation and
    corresponding FileStorage objects are the same.

    Also collect files and assignes them to corresponding FileStorage object.
    Assumption: all files at the locotion are collected.
    """
    DataLocation = apps.get_model("flow", "DataLocation")
    FileStorage = apps.get_model("storage", "FileStorage")
    StorageLocation = apps.get_model("storage", "StorageLocation")
    ReferencedPath = apps.get_model("storage", "ReferencedPath")

    max_id = 1
    for data_location in DataLocation.objects.all():
        file_storage = FileStorage.objects.create(id=data_location.id)
        base_dir = os.path.join(
            settings.FLOW_EXECUTOR["DATA_DIR"], data_location.subpath
        )
        files, dirs = collect(base_dir)
        ReferencedPath.objects.bulk_create(
            [ReferencedPath(path=path, file_storage=file_storage) for path in dirs]
        )
        ReferencedPath.objects.bulk_create(
            [
                ReferencedPath(path=path, file_storage=file_storage, size=size)
                for path, size in files
            ]
        )
        StorageLocation.all_objects.create(
            id=data_location.id,
            file_storage=file_storage,
            url=data_location.subpath,
            status="OK",
            connector_name=getattr(settings, "STORAGE_LOCAL_CONNECTOR", "local"),
        )
        max_id = max(max_id, data_location.id + 1)

    # Increment FileStorage and StorageLocation id's sequence
    if FileStorage.objects.exists():
        with connection.cursor() as cursor:
            cursor.execute(
                "ALTER SEQUENCE storage_filestorage_id_seq RESTART WITH {};".format(
                    max_id
                )
            )
            cursor.execute(
                "ALTER SEQUENCE storage_storagelocation_id_seq RESTART WITH {};".format(
                    max_id
                )
            )


class Migration(migrations.Migration):
    dependencies = [
        ("storage", "0001_initial"),
        ("flow", "0043_full_text_search"),
    ]

    run_before = [("flow", "0044_datalocation_to_filestorage")]

    operations = [
        migrations.RunPython(create_new_storage_locations),
    ]
